import json
from typing import Any, Callable, Dict, List, Tuple

from .. import errors as E
from ..color import color
from ..config import Config
from ..logger import logger
from .docker import DockerImage


class ImageGraph:
    def __init__(self, config: Config, image_factory: Callable) -> None:
        self.config: Config = config
        self.image_factory: Callable = image_factory
        self.images: Dict[str, DockerImage] = {}

    def configure_target_image(self, target: str) -> None:
        if target not in self.images:
            image = self.images[target] = self.image_factory(target)
            image.logger.info(f"target_image {color.image(image.name)}")

    def _resolve_target_dependencies(
        self, target: str, dependencies: List[str], vectors: List[Tuple[str, str]]
    ) -> List[str]:
        # use a depth-first search to unroll dependencies and detect loops
        for dependent in self.config.get_target(target).get("depends_on", []):
            vector = (target, dependent)
            if vector in vectors:
                raise E.TargetCyclicError(
                    f"Target [{target}] has circular dependencies"
                )
            else:
                vectors.append(vector)
                self._resolve_target_dependencies(dependent, dependencies, vectors)
                vectors.pop()

        self.configure_target_image(target)
        if target not in dependencies:
            dependencies.append(target)

        return dependencies

    def resolve_target_dependencies(self, target: str) -> List[DockerImage]:
        return self._resolve_target_dependencies(
            target=target, dependencies=[], vectors=[]
        )

    @property
    def allowed_pull_errors(self) -> Tuple[Any, ...]:
        allowed_error_names = self.config.allowed_pull_errors
        return tuple(getattr(E, name) for name in dir(E) if name in allowed_error_names)

    def pull_target_image(self, image: DockerImage) -> None:
        try:
            image.logger.info(f"{color.image(image.name)} Trying pull...")
            image.pull()
        except Exception as error:
            if not isinstance(error, self.allowed_pull_errors):
                raise
            else:
                error_name = type(error).__name__
                error_message = color.error(f"{error_name}: {error}")
                image.logger.info(f"{color.image(image.name)} {error_message}")
                image.logger.info(
                    f"{color.image(image.name)} {error_name} in allowed_pull_errors, continuing..."
                )

    def _load_target_image(self, image: DockerImage) -> bool:
        if image.is_loaded:
            image.logger.info(f"{color.image(image.name)} Image loaded")
        else:
            image.logger.info(f"{color.image(image.name)} Image not loaded")
            self.pull_target_image(image)

        if image.is_loaded and self.config.validate_image:
            image.validate()
            image.logger.info(f"{color.image(image.name)} Image validated")

        return image.is_loaded

    def load_target_image(self, image: DockerImage) -> bool:
        is_loaded = False

        try:
            is_loaded = self._load_target_image(image)
        except E.ImageValidationError as error:
            if not self.config.build_on_validate_error:
                raise
            else:
                error_message = color.error(f"ImageValidationError: {error}")
                image.logger.warning(f"{color.image(image.name)} {error_message}")
                image.logger.warning(
                    f"{color.image(image.name)} build_on_validate_error enabled, continuing..."
                )

        return is_loaded

    @staticmethod
    def display_image_build_options(image: DockerImage) -> None:
        json_text = json.dumps(image.build_options, sort_keys=True, indent=4)
        json_lines = json_text.splitlines()
        image.logger.info(
            f"{color.image(image.name)} build_options {color.config(json_lines.pop(0))}"
        )

        for line in json_lines:
            image.logger.info(color.config(line))

    def _build_image(self, image: DockerImage) -> None:
        image.logger.info(f"{color.image(image.name)} Trying build...")
        self.display_image_build_options(image)
        image.build()

        if self.config.prune_after_build:
            image.prune()

    def build_target_images(self, target: str, force: bool = False) -> None:
        image = self.images[target]

        if force or not self.load_target_image(image):
            for dependent in self.config.get_target(target).get("depends_on", []):
                self.build_target_images(dependent, force)
            self._build_image(image)

    def display_build_target(self, target: str) -> None:
        targets = self.resolve_target_dependencies(target)

        if len(targets) > 1:
            dependencies = color.header(" with dependencies: ")
            dependencies += color.header(", ").join(
                [color.target(f"[{target}]") for target in targets[-2::-1]]
            )
        else:
            dependencies = ""

        logger.info(
            "\n{0} {1} {2}{3}".format(
                color.header("Building"),
                color.target(f"[{target}]"),
                color.header("target"),
                dependencies,
            )
        )

    def resolve_build_targets(self, targets: List[str]) -> None:
        logger.info(color.header("\nResolving build targets"))
        for target in targets:
            self.resolve_target_dependencies(target)

    def build(self, targets: List[str], force: bool = False) -> None:
        self.resolve_build_targets(targets)

        for target in targets:
            self.display_build_target(target)
            self.build_target_images(target, force)

    def push(self, targets: List[str]) -> None:
        logger.info(color.header("Pushing images"))

        for image in [self.images[target] for target in targets]:
            if image.is_loaded:
                if image.was_pulled and not image.was_built:
                    image.logger.info(
                        f"{color.image(image.name)} Skipping push of pulled image"
                    )
                else:
                    image.logger.info(f"{color.image(image.name)} Trying push...")
                    image.push()



import argparse
import datetime
import json
import sys
import time
import traceback
from typing import Dict, List, Tuple

from . import __version__ as VERSION
from . import constants as C
from . import envfile
from .color import color
from .config import Config
from .logger import logger
from .image import DockerImage, FakeImage, ImageFactory, ImageGraph


def _arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Build container images faster ⚡️")

    parser.add_argument(
        "--config",
        "-c",
        dest="config_file",
        action="store",
        metavar="PATH",
        default=C.DEFAULT_CONFIG_FILE,
        help="Target build config file to use (default: %(default)s)",
    )
    parser.add_argument(
        "--dry-run",
        dest="noop",
        action="store_true",
        help="Show how targets would be built (implies --force)",
    )
    parser.add_argument(
        "--force",
        dest="force",
        action="store_true",
        help="Force target(s) to be rebuilt",
    )
    parser.add_argument(
        "--monochrome",
        "-m",
        dest="color",
        action="store_false",
        help="Don't colorize log messages",
    )
    parser.add_argument(
        "--push",
        "-p",
        nargs="+",
        metavar="TARGET_NAME",
        help="Image target(s) to push to repo after building",
    )
    parser.add_argument(
        "--repo",
        "-r",
        dest="repo",
        action="store",
        metavar="URI",
        required=True,
        help="Image repo to use (e.g. quay.io/lexsca/prefab)",
    )
    parser.add_argument(
        "--target",
        "-t",
        dest="_targets",
        nargs="+",
        required=True,
        metavar="TARGET_NAME[:TAG]",
        help="Image target(s) to build with optional custom image tag",
    )

    return parser


def parse_options(args: List[str]) -> argparse.Namespace:
    parser = _arg_parser()
    options = parser.parse_args(args)
    options.tags = dict()
    options.targets = []
    tag_set = set()

    for target in options._targets:
        target, _, tag = target.partition(":")
        options.targets.append(target)

        if tag and tag in tag_set:
            parser.exit(
                status=2, message=f"{parser.format_help()}\nDuplicate tag: {tag}\n"
            )
        elif tag:
            options.tags[target] = tag
            tag_set.add(tag)

    if options.noop:
        options.force = True

    return options


def parse_args(args: List[str]) -> Tuple[argparse.Namespace, Config]:
    options = parse_options(args)
    config = Config.from_yaml_filepath(options.config_file)

    if not options.color:
        color.enabled = False
    else:
        color.style.update(config.color_style)

    logger.info(color.header(f"\nContainer Prefab v:{VERSION}"))
    logger.info(f"Called with args: {color.config(json.dumps(args))}")
    logger.info(f"Loaded config file: {color.config(config.path)}")
    write_envfiles()

    logger.info(color.header("\nConfig options:"))
    config.display_options()

    return options, config


def write_envfiles(envfiles: Dict[str, str] = C.DEFAULT_ENVFILES) -> None:
    for env_key, dest_path in envfiles.items():
        envfile.write(env_key, dest_path)


def elapsed_time(monotonic_start: float) -> str:
    elapsed_time = datetime.timedelta(seconds=time.monotonic() - monotonic_start)
    seconds = elapsed_time.seconds
    category = 1 + (-1 if seconds < 10 else seconds // 60 if seconds < 240 else 4)
    awards = ("⚡", "🚀", "🥇", "🥈", "🥉", "🐢")

    return f"{str(elapsed_time)[:-4]} {awards[category]}"


def cli(args: List[str]) -> None:
    build_start_time = time.monotonic()
    options, config = parse_args(args)

    image_constructor = FakeImage if options.noop else DockerImage
    image_factory = ImageFactory(config, options.repo, options.tags, image_constructor)
    image_graph = ImageGraph(config, image_factory)
    image_graph.build(options.targets, options.force)
    logger.info(
        f"\n{color.elapsed('Build elapsed time:')} {elapsed_time(build_start_time)}\n"
    )

    if options.push:
        push_start_time = time.monotonic()
        image_graph.push(options.push)
        logger.info(
            f"\n{color.elapsed('Push elapsed time: ')} {elapsed_time(push_start_time)}"
        )
        logger.info(
            f"{color.elapsed('Total elapsed time:')} {elapsed_time(build_start_time)}\n"
        )


def main() -> None:
    try:
        cli(sys.argv[1:])
        status = 0
    except Exception:
        logger.error(color.error(traceback.format_exc()))
        status = 1
    sys.exit(status)



from dataclasses import dataclass
from typing import Optional

from .docker import DockerImage


@dataclass
class RaiseOn:
    pull: Optional[Exception]
    validate: Optional[Exception]
    build: Optional[Exception]
    push: Optional[Exception]


class FakeImage(DockerImage):
    def __init__(
        self,
        *args,
        loaded=False,
        pull=None,
        validate=None,
        build=None,
        push=None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.raise_on = RaiseOn(pull, validate, build, push)
        self._loaded = loaded

    def pull(self) -> None:
        if self.raise_on.pull:
            raise self.raise_on.pull

        self._pull_success()

    def validate(self) -> None:
        if self.raise_on.validate:
            raise self.raise_on.validate

    def build(self) -> None:
        if self.raise_on.build:
            raise self.raise_on.build

        self._build_success()

    def push(self) -> None:
        if self.raise_on.push:
            raise self.raise_on.push

        self.logger.info(f"{self.name} Pushed")



import fnmatch
import os
from typing import List, Tuple


def parse_ignore_patterns(patterns: List[str]) -> Tuple[List[str], List[str]]:
    file_patterns: List[str] = []
    dir_patterns: List[str] = []

    for pattern in patterns:
        if pattern.endswith("/"):
            dir_patterns.append(pattern[:-1])
        else:
            dir_patterns.append(pattern)
            file_patterns.append(pattern)

    return dir_patterns, file_patterns


def cull_dirs(
    base_dir: str,
    dir_names: List[str],
    file_names: List[str],
    ignore_patterns: List[str],
) -> None:
    ignore_dirs = set()

    for pattern in ignore_patterns:
        if fnmatch.fnmatch(base_dir, pattern):
            dir_names.clear()
            file_names.clear()
            ignore_dirs.clear()
            break

        for dir_name in fnmatch.filter(dir_names, pattern):
            ignore_dirs.add(dir_name)

    for dir_name in ignore_dirs:
        dir_names.remove(dir_name)


def filter_files(
    base_dir: str, file_names: List[str], ignore_patterns: List[str]
) -> List[str]:
    file_paths = []
    ignore_names = set()

    for ignore_pattern in ignore_patterns:
        for file_name in fnmatch.filter(file_names, ignore_pattern):
            ignore_names.add(file_name)

    for file_name in set(file_names) - ignore_names:
        file_paths.append(os.path.join(base_dir, file_name))

    for pattern in ignore_patterns:
        for file_path in fnmatch.filter(file_paths, pattern):
            file_paths.remove(file_path)

    return file_paths


def walk(path: str, ignore_patterns: List[str]) -> List[str]:
    file_paths = []
    norm_path = os.path.normpath(path)
    dir_ignore_patterns, file_ignore_patterns = parse_ignore_patterns(ignore_patterns)

    def onerror(error):
        raise error

    for base_dir, dir_names, file_names in os.walk(
        norm_path, topdown=True, onerror=onerror
    ):
        cull_dirs(base_dir, dir_names, file_names, dir_ignore_patterns)
        file_paths.extend(filter_files(base_dir, file_names, file_ignore_patterns))

    return sorted(file_paths)



import base64
import binascii
import os
import tempfile

from .color import color
from .logger import logger


def write(env_key: str, dest_path: str, dockerenv_path: str = "/.dockerenv") -> None:
    if os.path.exists(dockerenv_path) and env_key in os.environ:
        _write(env_key, dest_path)


def _write(env_key: str, dest_path: str) -> None:
    try:
        content = base64.b64decode(os.environ.get(env_key, ""))
        if content:
            logger.info(f"Writing env {env_key} to {color.config(dest_path)}")
            return _commit(dest_path, content)
        else:
            logger.warning(
                color.warning(f"Not writing empty env {env_key} to {dest_path}")
            )
    except binascii.Error as error:
        logger.warning(
            color.warning(
                f"Not writing env {env_key} to {dest_path}, base64 decode failed: {error}"
            )
        )
    except Exception as error:
        logger.warning(color.warning(f"{type(error).__name__}: {error}"))


def _commit(dest_path: str, content: bytes) -> None:
    dest_dirname = os.path.dirname(os.path.abspath(dest_path))
    temp_fp, temp_path = tempfile.mkstemp(dir=dest_dirname)

    with open(temp_fp, "wb") as temp:
        temp.write(content)

    os.rename(temp_path, dest_path)



from typing import Dict

from . import constants as C


class Color:
    reset: str = "\x1b[0m"

    def __init__(self, style: Dict[str, int] = C.DEFAULT_COLOR_STYLE) -> None:
        self.enabled: bool = True
        self.style: Dict[str, int] = style.copy()

        cls = type(self)

        for style in self.style:
            if not hasattr(cls, style):

                def colorizer(this, text: str, style: str = style) -> str:
                    # https://en.wikipedia.org/wiki/ANSI_escape_code#8-bit
                    code = f"\x1b[38;5;{self.style.get(style, 0)}m"
                    return f"{code}{text}{self.reset}" if this.enabled else text

                setattr(cls, style, colorizer)


color = Color()



class PrefabError(Exception):
    pass


class HashAlgorithmNotFound(PrefabError):
    pass


class ImageAccessError(PrefabError):
    pass


class ImageBuildError(PrefabError):
    pass


class ImageNotFoundError(PrefabError):
    pass


class ImagePushError(PrefabError):
    pass


class ImageValidationError(PrefabError):
    pass


class InvalidConfigError(PrefabError):
    pass


class TargetCyclicError(PrefabError):
    pass


class TargetNotFoundError(PrefabError):
    pass
